"""
Python Currency Converter CLI
Official Repo: https://github.com/Privex/python-curconv
License: X11 / MIT


Copyright::

    +===================================================+
    |                 © 2021 Privex Inc.                |
    |               https://www.privex.io               |
    +===================================================+
    |                                                   |
    |        Python Currency Converter CLI              |
    |        License: X11/MIT                           |
    |                                                   |
    |        Core Developer(s):                         |
    |                                                   |
    |          (+)  Chris (@someguy123) [Privex]        |
    |                                                   |
    +===================================================+
    
    Python Currency Converter CLI - A small CLI tool for quick currency conversions on the command line
    Copyright (c) 2021    Privex Inc. ( https://www.privex.io )
    
    Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:
    
    The above copyright notice and this permission notice shall be included in all copies or substantial portions of
    the Software.
    
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
    WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
    OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
    
    Except as contained in this notice, the name(s) of the above copyright holders shall not be used in advertising or
    otherwise to promote the sale, use or other dealings in this Software without prior written authorization.

"""
import logging
# logging.basicConfig(level=logging.DEBUG, force=True)
name = 'curconv'
from privex.curconv.version import VERSION

VERSION = VERSION
from privex.curconv import settings
from privex.curconv.app import main, convert, str_convert, str_extract, conv_adapter
from privex.curconv.base import *



import json
import os
import sys
import logging
import dotenv
from os import getenv as env
from pathlib import Path
from decimal import getcontext
from privex.helpers import empty, env_int, env_bool, env_csv, env_keyval,  HOUR, DictObject
from privex.helpers import settings as pvx_settings
from privex.helpers.cache import async_adapter_set, adapter_set
from privex.loghelper import LogHelper

DEBUG = env_bool('DEBUG', False)

CONF_PATH = env_csv('CONF_PATH', ['~/.curconv', '~/.curconv.env', '/etc/curconv', '/etc/curconv.env'], csvsplit=":")
"""
To override the list of config path locations, the env var CONF_PATH has to be set
to a colon separated list of paths, similar to the system PATH var.
Example: ``~/.curconv:~/.curconv.env:/etc/curcon``
"""

CONF_OVERRIDE = env_bool('CONF_OVERRIDE', False)
"""
If set to ``True`` in the environment, then variables in the .env config file will take
priority over system environment vars. Otherwise, .env vars won't override system ones.
"""

# CONF_LOCAL_PATH_ALT = env('CONF_LOCAL_PATH', '~/.curconv.env')
CONF_LOCAL = None


########
# Iterate through CONF_PATH to try and locate a .env config file which exists.
########
for p in CONF_PATH:
    lp = Path(p).expanduser().resolve()
    if lp.exists():
        if DEBUG: print(f"Found existing file in CONF_PATH - using as CONF_LOCAL: {lp!s}", file=sys.stderr)
        CONF_LOCAL = lp
        break
    if DEBUG: print(f"SKIPPING - Path does not exist: {lp!s}", file=sys.stderr)

if not empty(CONF_LOCAL):
    try:
        # First assume dotenv = python-dotenv
        dotenv.load_dotenv(str(CONF_LOCAL), override=CONF_OVERRIDE)
    except (AttributeError, KeyError, IndexError):
        # If load_dotenv throws certain exceptions, it may be due to
        # dotenv = django-dotenv, instead of python-dotenv, so try read_dotenv instead.
        dotenv.read_dotenv(str(CONF_LOCAL), override=CONF_OVERRIDE)

# Now that we've potentially loaded a .env file, re-set DEBUG incase it's influenced by .env
# print("__name__ is: ", __name__)
# print(json.dumps(dict(os.environ), indent=4))
DEBUG = env_bool('DEBUG', False)
CONF_DEFAULT = Path('~/.curconv.env').expanduser().resolve()
LOG_LEVEL = env('LOG_LEVEL', 'DEBUG' if DEBUG else 'WARNING').upper()

# _lh = LogHelper('privex.curconv', handler_level=logging.getLevelName(LOG_LEVEL), clear_handlers=False)
_lh = LogHelper('privex.curconv', handler_level=logging.getLevelName(LOG_LEVEL))
_lh.copy_logger('privex')
h = _lh.add_console_handler(stream=sys.stderr)

# _lh.copy_logger('privex.curconv.base', 'privex.curconv.ratesapi', 'privex.curconv.app', 'conv')

# log = _lh.get_logger()
# log.propagate = True
getcontext().prec = 28

# API_BASE = env('API_BASE', 'https://api.exchangeratesapi.io').rstrip('/')
API_BASE = env('API_BASE', 'https://api.ratesapi.io/api').rstrip('/')
API_URI = env('API_URI', '/latest')
API_CACHE_TIME = env_int('API_CACHE_TIME', 1 * HOUR)
pvx_settings.SQLITE_APP_DB_NAME = env('SQLITE_APP_DB_NAME', 'pvx-curconv')
CACHE_ADAPTER = env('CACHE_ADAPTER', 'sqlite')
"""
Available adapters:

  * ``memory`` - Not really useful for this kind-of application, as the cache is stored entirely
                 in the memory of the app, thus is lost as soon as the command finishes.
  * ``sqlite`` - Stores the cache in an SQLite3 database. By default, the database will be stored
                 at ``~/.privex_cache/{SQLITE_APP_DB_NAME}``
                 If you haven't changed :attr:`.SQLITE_APP_DB_NAME` then it'll be stored in
                 the database ``~/.privex_cache/pvx-curconv``
  * ``redis``  - (Recommended) -
                 
"""

CUR_SYMBOLS = {
    'CA$':  'CAD',
    'CAD$': 'CAD',
    'AU$':  'AUD',
    'AUD$': 'AUD',
    'NZ$':  'NZD',
    'NZD$': 'NZD',
    '$':    'USD',
    '€':    'EUR',
    '£':    'GBP',
    'kr':   'SEK',
}

CUST_CUR_SYMBOLS = dict(env_keyval('CUR_SYMBOLS', {}))
CUR_SYMBOLS = {**CUR_SYMBOLS, **CUST_CUR_SYMBOLS}

CONF = DictObject(
    dp=env_int('DEFAULT_DP', 3),
    base_coin=env('BASE_COIN', 'EUR'),
    cur_list=env_csv('CURRENCY_LIST', ['USD', 'EUR', 'GBP', 'CAD', 'SEK', 'TRY'])
)

adapter_set(CACHE_ADAPTER)
async_adapter_set(CACHE_ADAPTER)



#!/usr/bin/env python3
"""
High level functions/classes for running the Python Currency Converter CLI

Official Repo: https://github.com/Privex/python-curconv
License: X11 / MIT


Copyright::

    +===================================================+
    |                 © 2021 Privex Inc.                |
    |               https://www.privex.io               |
    +===================================================+
    |                                                   |
    |        Python Currency Converter CLI              |
    |        License: X11/MIT                           |
    |                                                   |
    |        Core Developer(s):                         |
    |                                                   |
    |          (+)  Chris (@someguy123) [Privex]        |
    |                                                   |
    +===================================================+

    Python Currency Converter CLI - A small CLI tool for quick currency conversions on the command line
    Copyright (c) 2021    Privex Inc. ( https://www.privex.io )

"""
import sys
import asyncio
import textwrap
import argparse
from pathlib import Path
from decimal import Decimal
from privex.helpers import dec_round, ErrHelpParser
from typing import Union
from privex.curconv import settings
from privex.curconv.base import gen_env, is_numstr, replace_sym, to_keyval, trans_dict, trans_rate
from privex.curconv.ratesapi import RatesAPIAdapter
from privex.curconv.version import VERSION
import logging

# print("__name__ is: ", __name__)

__all__ = [
    'conv_adapter', 'convert', 'str_extract', 'str_convert',
    'main',
]

log = logging.getLogger(__name__)

conv_adapter = RatesAPIAdapter()
CONF = settings.CONF


async def convert(
    amt: Union[Decimal, float, int, str], from_coin: str, to_coin: str, base_coin: str = None
):
    if not isinstance(amt, Decimal):
        amt = Decimal(amt)
    log.debug(f" [convert] Calling conv_adapter.get_rates for base coin: {base_coin!r}")
    rates = await conv_adapter.get_rates(base=base_coin)
    trate = trans_rate(from_coin, to_coin, rates)
    return amt * trate


def str_extract(query: str):
    sq = list(query.split())
    if not is_numstr(sq[0]):
        rsm = replace_sym(sq[0]).split()
        if len(rsm) < 2:
            raise ValueError(f"First segment of string isn't a number, nor has a valid symbol prefix/suffix. Segment is: {sq[0]!r}")
        sq[0] = rsm[0]
        sq.insert(1, rsm[1])
    amt = Decimal(sq[0])
    fcoin = sq[1]
    log.debug(f" [str_extract] Extracted amt + fcoin: amt = {amt!r}; fcoin = {fcoin!r}")

    if len(sq) <= 2:
        return [amt, fcoin]
    dcoin = sq[3]
    if dcoin in settings.CUR_SYMBOLS:
        dcoin = replace_sym(dcoin)
    log.debug(f" [str_extract] Full extraction. amt = {amt!r}; fcoin = {fcoin!r}; dcoin = {dcoin!r}")

    return [amt, fcoin, dcoin]


async def str_convert(query: str, base_coin='EUR'):
    sq = str_extract(query)
    log.debug(f" [str_convert] Calling conv_adapter.get_rates with base_coin: {base_coin!r}")
    
    rates = await conv_adapter.get_rates(base_coin)
    amt, fcoin = sq[0], sq[1]
    if len(sq) <= 2:
        if fcoin.upper() == base_coin.upper():
            return {k: amt * v for k, v in rates['rates'].items()}
        tdt = trans_dict(fcoin, rates)
        return {k: amt * v for k, v in tdt.items()}
    dcoin = sq[2]
    if fcoin.upper() == base_coin.upper():
        return amt * rates['rates'].get(dcoin, rates['rates'].get(dcoin.upper()))
    return amt * trans_rate(fcoin, dcoin, rates)


async def main():
    descr = textwrap.dedent(f"""
        CLI Currency Converter - Version v{VERSION}
        (C) 2021 Someguy123 ( https://github.com/Someguy123 )
            Privex Inc.     ( https://www.privex.io )
            Official Repo:  https://github.com/Privex/python-curconv

        Easily get the exchange rate / conversion amount for either an individual coin pair,
        for any currency into all of your favourite currencies, or for any currency into
        all other currencies.

        Examples:

            # Show how much 1 USD would get you in your favourite currencies
            $ {sys.argv[0]} 1 usd
            # Show how much 1000 SEK is worth in USD
            $ {sys.argv[0]} 1000 SEK to USD
            # Show how much 50 GBP is worth in EUR, with full mode enabled (show from_coin and to_coin in output)
            $ {sys.argv[0]} -f £50 to eur
            # Show how much 20 CAD is worth in all supported currencies, with 6 dp precision used for displaying exchange rates
            $ {sys.argv[0]} -d 6 -a CA$20
            # Show how much 50 AUD is worth in NZD, USD, EUR, and JPY
            $ {sys.argv[0]} -c nzd,usd,eur,jpy AU$50

        ------------

        Customization:

            Default ENV config search path (CONF_PATH - colon separated):
                {' : '.join(settings.CONF_PATH)}

            # Generate a template ENV config file to the default location: {settings.CONF_DEFAULT!s}
            $ {sys.argv[0]} -Gd

            # Generate a template ENV config to STDOUT (the terminal), for piping or copying
            $ {sys.argv[0]} -G
            $ {sys.argv[0]} -G | sudo tee /etc/curconv.env

            # Generate a template ENV config to the file /etc/curconv.env
            $ {sys.argv[0]} -G -Gf /etc/curconv.env

        ENV Vars:

            DEBUG (bool)            - Enables debug mode, which also boosts LOG_LEVEL to DEBUG. Default: 'false'
            LOG_LEVEL (str)         - The level of logging to output (DEBUG,INFO,WARNING,ERROR,CRITICAL). Default: {settings.LOG_LEVEL!r}
            API_BASE (str)          - The base API URL to use - Default: {settings.API_BASE!r}
            API_CACHE_TIME (int)    - The number of seconds to cache exchange rates - Default: {settings.API_CACHE_TIME}
            CACHE_ADAPTER (str)     - Adapter to use for caching: memory,redis,memcached - Default: {settings.CACHE_ADAPTER!r}
            DEFAULT_DP (int)        - Decimal places to round output numbers to - Default: {settings.CONF.dp}
            BASE_COIN (str)         - The base coin used by the API for conversion - Default: {settings.CONF.base_coin!r}
            CURRENCY_LIST (str)     - Comma separated list of "favourite" currency codes.
                                      Default: {','.join(settings.CONF.cur_list)!r}
            CUR_SYMBOLS (str)       - A key value map of symbols to currencies, e.g. '$:USD,£:GBP,kr:SEK' etc.
                                      Default: {to_keyval(settings.CUR_SYMBOLS)!r}

            CONF_DEFAULT (str)      - This setting isn't too important, it purely controls where the
                                      "default config file" is considered to be located, and is only
                                      currently used when the '-Gd' flag is passed, to generate a
                                      template .env config and output it to the default location.
                                      Default: {str(settings.CONF_DEFAULT)!r}


        System Environment ONLY:

            These are ENV vars which can only be set dynamically by either injecting them with your shell
            when running the CLI tool, or exporting them before running the tool. They can't be set in
            the config (.env) file due to them being used prior to loading the .env file.

            (partially) DEBUG (bool)    - While DEBUG can be set in your config file, the config setting only
                                          affects the code ran AFTER your .env is loaded. If you set DEBUG=true
                                          in the system environment when running the tool, it enables
                                          debug logging early-on, including during the code which actually
                                          locates an existing .env file and loads it.

            CONF_OVERRIDE (bool)        - This is a boolean setting (Default: False) which controls whether
                                          settings in your .env config file take priority over system env vars.
                                          When True, if you pass 'DEBUG=true' via system env vars, but your
                                          .env config contains 'DEBUG=false', then your config setting takes
                                          priority and DEBUG will be false.
                                          When False, in the above scenario, the system env var would take
                                          priority and DEBUG would be True.

            CONF_PATH (str)             - This variable controls the list of config file locations to check
                                          for an existing file, in priority order (leftmost first).
                                          It's read similar to the system PATH variable, split into a
                                          list using the ':' colon character as the separator,
                                          e.g. '~/.curconv:~/.curconv.env:/etc/curcon'.
                                          The default value for this is: {':'.join(settings.CONF_PATH)!r}

    """)
    
    parser = ErrHelpParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=descr
    )
    parser.add_argument('-d', '-dp', '--dp', '--decimals', '--decimal-places', default=CONF.dp, dest='dp',
                        help="The number of decimal places to round rate results to. Default: {CONF.dp}")
    parser.add_argument('-b', '--base', '--base-coin', default=CONF.base_coin, dest='base_coin',
                        help=f"Normally you don't need to change this. This controls the 'base' or 'root' currency "
                             f"which all rates from the API are in, which is used to inversely calculate the "
                             f"exchange rate between two or more arbitrary currencies. Default: {CONF.base_coin!r}")
    parser.add_argument('-c', '--currencies', default=','.join(CONF.cur_list), dest='currencies',
                        help="Use '-c' to override the default 'favourite'/'selected' currency list "
                             "with the currencies you actually want to display, e.g. '-c USD,AUD,NZD,TRY'")
    parser.add_argument('-a', '-ac', '--all-currencies', default=False, action='store_true', dest='all_currencies',
                        help="When querying just the source i.e. '10.32 USD', the '-a' flag results in "
                             "all supported currencies exchange rates being outputted, instead of just "
                             "favourited/selected currencies.")
    parser.add_argument('-f', '--full', default=False, action='store_true', dest='full_mode',
                        help="In full mode, the source amount and currency will be displayed with "
                             "an '=' equals sign before the destination(s), and for direct single "
                             "conversions, it enables displaying the destination currency too, "
                             "e.g. query '50 GBP to EUR' outputs '57.61' normally, but in "
                             "full mode it would output '50 GBP = 57.61 EUR'.")
    parser.add_argument('-G', '-genv', '--gen-env', '--generate-env', default=False,
                        action='store_true', dest='gen_env',
                        help=f"Generate a .env file from current loaded settings. Outputs to stdout "
                             f"unless you pass either '-Gf [file]' to output to a file, or '-Gd' "
                             f"to output to the default config location {str(settings.CONF_DEFAULT)!r}")
    
    parser.add_argument('-Gf', '--env-out', '--gen-env-out', default="-", dest='env_out',
                        help="Set the output location for the .env generated when used with '-G'")
    parser.add_argument('-Gd', '--env-default', '--gen-env-default', default=False,
                        action='store_true', dest='env_out_default',
                        help=f"Generate and output a .env file from current loaded settings "
                             f"to the default config location: {str(settings.CONF_DEFAULT)!r}")
    parser.add_argument('query', nargs='*', help="The conversion query, e.g. '4.20 USD to GBP' or '£5.20 to EUR'")
    zargs = parser.parse_args()
    z_gen = zargs.gen_env or zargs.env_out_default
    
    if z_gen:
        xenv = gen_env()
        if zargs.env_out_default:
            print(f" >>> Generating ENV file to default location: {settings.CONF_DEFAULT}", file=sys.stderr)
            with open(str(settings.CONF_DEFAULT), 'w') as fh:
                fh.write(xenv)
            return print(f" +++ Successfully wrote ENV file to default location: {settings.CONF_DEFAULT}", file=sys.stderr)
        
        if zargs.env_out in ['', '-', '/dev/stdout', 'stdout', 'STDOUT']:
            return print(xenv)
        
        if zargs.env_out in ['/dev/stderr', 'stderr', 'STDERR']:
            return print(xenv, file=sys.stderr)
        out_loc = Path(zargs.env_out).expanduser().resolve()
        print(f" >>> Generating ENV file to custom location: {out_loc!s}", file=sys.stderr)
        with open(str(out_loc), 'w') as fh:
            fh.write(xenv)
        return print(f" +++ Successfully wrote ENV file to custom location: {out_loc!s}", file=sys.stderr)
    
    CONF.dp = zargs.dp
    CONF.base_coin = zargs.base_coin.upper()
    CONF.cur_list = zargs.currencies.split(',')
    if len(zargs.query) == 0:
        return parser.error("No query specified! You must specify a query e.g. '5 usd to gbp' or '2.1 cad'")
    q = ' '.join(zargs.query)
    
    log.debug(f" [main] Entering conversion adapter using 'async with'")
    async with conv_adapter:
        log.debug(f" [main] Calling str_convert with arguments: (0 'q') = {q!r} ; base_coin = {CONF.base_coin!r} ;")
        sv = await str_convert(q, base_coin=CONF.base_coin)
        log.debug(f" [main] Calling str_extract with arguments: (0 'q') = {q!r} ;")
        sq = str_extract(q)
        amt, fcoin = sq[0], sq[1]
    
    if isinstance(sv, dict):
        if zargs.full_mode:
            print(f"  {amt} {fcoin.upper()} =\n")
        rsv = {k: dec_round(v, CONF.dp) for k, v in sv.items()}
        if zargs.all_currencies:
            for k, v in rsv.items():
                if k.upper() == fcoin.upper(): continue
                print(f"{v} {k.upper()}")
        else:
            for k in CONF.cur_list:
                if k.upper() == fcoin.upper(): continue
                print(f"{rsv[k.upper()]} {k.upper()}")
        print()
        return
    dcoin = sq[2]
    sv = dec_round(sv, CONF.dp)
    if zargs.full_mode:
        print(f"{amt} {fcoin.upper()} = {sv} {dcoin.upper()}")
    else:
        print(f"{sv}")


if __name__ == '__main__':
    asyncio.run(main())





#!/usr/bin/env python3
import asyncio
import sys
from pathlib import Path
loc = Path(__file__).expanduser().resolve().parent
deps_dir = (loc / 'deps').resolve()

##########
# The following code is for the .pyz self-contained executable version
# of privex-curconv - it adds the .pyz virtual folder to the start of
# the Python search path, to allow both the app and modules to be loaded.
#
if str(loc).rstrip('/').endswith('.pyz') or str(loc).rstrip('/').endswith('.zip'):
    # print(f"loc = {str(loc)}")
    # print(f"deps_dir = {str(deps_dir)}")
    # print(f"sys.path = {sys.path!r}")
    sys.path.insert(0, str(loc))
    # print(f"(inserted) sys.path = {sys.path!r}")
    # sys.path.append(str(deps_dir))


from privex.curconv.app import main

if __name__ == '__main__':
    asyncio.run(main())




"""
RatesAPI.io Currency Adapter for the Python Currency Converter CLI

Official Repo: https://github.com/Privex/python-curconv
License: X11 / MIT


Copyright::

    +===================================================+
    |                 © 2021 Privex Inc.                |
    |               https://www.privex.io               |
    +===================================================+
    |                                                   |
    |        Python Currency Converter CLI              |
    |        License: X11/MIT                           |
    |                                                   |
    |        Core Developer(s):                         |
    |                                                   |
    |          (+)  Chris (@someguy123) [Privex]        |
    |                                                   |
    +===================================================+

    Python Currency Converter CLI - A small CLI tool for quick currency conversions on the command line
    Copyright (c) 2021    Privex Inc. ( https://www.privex.io )

"""
import json
import httpx
from decimal import Decimal
from privex.helpers import empty_if
from privex.curconv import settings
from privex.curconv.base import CurrencyAdapter, Pair, PairList
import logging

__all__ = ['RatesAPIAdapter']

# log = logging.getLogger('privex.curconv')
log = logging.getLogger(__name__)
# log.setLevel('DEBUG')


# log.propagate = True
# log.


class RatesAPIAdapter(CurrencyAdapter):
    DEFAULT_CACHE_PREFIX = "rates_api:conv"
    
    def __init__(self, base_coin: str = settings.CONF.base_coin, **kwargs):
        kwargs = dict(kwargs)
        conf = dict(
            api_base=kwargs.pop('api_base', settings.API_BASE),
            api_uri=kwargs.pop('api_uri', settings.API_URI),
        )
        conf = {**kwargs, **conf}
        super().__init__(
            base_coin, **conf
        )
        if 'http2' not in self.httpx_conf: self.httpx_conf['http2'] = True
        if 'timeout' not in self.httpx_conf: self.httpx_conf['timeout'] = 10
    
    async def _get_rates(self, base: str = None, *symbols, **kwargs) -> PairList:
        symbols = list(symbols)
        if len(symbols) == 1 and "," in symbols[0]:
            symbols = [s.strip() for s in list(symbols[0].split(','))]
        base = empty_if(base, self.base_coin).upper()
    
        raise_status = kwargs.pop('raise_status', kwargs.pop('raise_for_status', True))
    
        data = dict(base=base)
        if len(symbols) > 0:
            data['symbols'] = ','.join(symbols)
    
        uri = f"{self.api_base}{self.api_uri}"
    
        log.info(f"Retrieving exchange rates from {uri!r} - base coin: {base!r} || symbols: {symbols!r}")
        # async with httpx.AsyncClient(http2=True, timeout=10) as h:
        async with self.httpx as h:
            res = await h.get(uri, params=data)
            if raise_status:
                res.raise_for_status()
            dc = ""
            async for ta in res.aiter_text():
                dc += ta
        log.debug(f"Decoding exchange rate JSON: {dc!r}")
        j = json.loads(dc, parse_float=Decimal)
    
        if 'rates' in j:
            if base not in j['rates']:
                j['rates'][base] = Decimal('1.0000000')
        
        xpairs = []
        for curr, rate in j['rates'].items():
            xpairs.append(Pair(base, curr, rate))
        return PairList(pairs=xpairs)



"""
Common/shared functions/classes/objects used by various modules in the Python Currency Converter CLI

Official Repo: https://github.com/Privex/python-curconv
License: X11 / MIT


Copyright::

    +===================================================+
    |                 © 2021 Privex Inc.                |
    |               https://www.privex.io               |
    +===================================================+
    |                                                   |
    |        Python Currency Converter CLI              |
    |        License: X11/MIT                           |
    |                                                   |
    |        Core Developer(s):                         |
    |                                                   |
    |          (+)  Chris (@someguy123) [Privex]        |
    |                                                   |
    +===================================================+

    Python Currency Converter CLI - A small CLI tool for quick currency conversions on the command line
    Copyright (c) 2021    Privex Inc. ( https://www.privex.io )

"""
import abc
import asyncio
import contextlib
import time
from dataclasses import dataclass, field
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple, Union, AsyncContextManager

import httpx
from async_property import async_property
from privex.helpers import AsyncCacheAdapter, DictDataClass, await_if_needed, empty
from privex.helpers.cache import AsyncCacheWrapper, async_cached

from privex.curconv import settings

import logging

__all__ = [
    'replace_sym', 'sort_dict', 'dump_args', 'to_keyval', 'gen_env',
    '_gen_env', 'Pair', 'AnyPair', 'IterAnyPair', 'PairList', 'CurrencyAdapter',
    'trans_rate', 'trans_dict', 'is_numstr'
]

log = logging.getLogger(__name__)


def replace_sym(data: str) -> str:
    """
    Converts currency strings such as ``£5.00`` to ``5.00 GBP`` - or ``10 kr`` to ``10 SEK``
    """
    origdata = data
    data = data.strip()
    for s, r in settings.CUR_SYMBOLS.items():
        if data.startswith(s) or data.endswith(s):
            log.debug(f"Replacing symbol {s!r} with {r!r}")
            return f"{data.replace(s, '').strip()} {r}".strip()
        if data.upper().startswith(s) or data.upper().endswith(s):
            log.debug(f"Replacing symbol {s!r} with {r!r} (uppercase)")
            return f"{data.upper().replace(s, '').strip()} {r}".strip()
    return origdata


def sort_dict(data: dict) -> dict:
    data = dict(data)
    skeys = sorted(list(data.keys()))
    return {k: data[k] for k in skeys}


def dump_args(*args, _serial_func=repr, **kwargs) -> str:
    """
    Convert positional args and keyword args into a flat string e.g. for cache keys.
    Sorts the keys in kwargs to ensure different ordering of kwargs (e.g. ``a='hello', b=123`` and ``b=123, a='hello'``)
    which have the same value, will still result in the same string.

    Example::

        >>> dump_args('lorem', 'ipsum', hello='world', donkey=420, banana=Decimal('1.23123'))
        "0='lorem';1='ipsum';banana=Decimal('1.23123');donkey=420;hello='world';"
        >>> dump_args('lorem', 'ipsum', donkey=420, hello='world', banana=Decimal('1.23123'), _serial_func=str)
        '0=lorem;1=ipsum;banana=1.23123;donkey=420;hello=world;'

    """
    
    args, kwargs = list(args), sort_dict(kwargs)
    dt = ""
    for i, a in enumerate(args):
        dt += f"{i}={_serial_func(a)};"
    for k, v in kwargs.items():
        dt += f"{k}={_serial_func(v)};"
    return dt


def to_keyval(ob: Union[dict, List[Tuple[Any, Any]]], valsplit=':', csvsplit=',') -> str:
    """
    Convert a :class:`.dict` , or a :class:`.list` of two value :class:`.tuple` 's
    into a "key value" string, which can be parsed by :func:`privex.helpers.common.parse_keyval`
    or manually with a small amount of python.

    Example::

        >>> d = dict(hello='world', lorem=123, ipsum='dolor')
        >>> to_keyval(d)
        'hello:world,lorem:123,ipsum:dolor'
        >>> to_keyval(d, valsplit='=', csvsplit='/')
        hello=world/lorem=123/ipsum=dolor

    """
    xo = ob.items() if isinstance(ob, dict) else ob
    return csvsplit.join([f'{k}{valsplit}{v}' for k, v in xo])


def _gen_env(**kwargs) -> dict:
    # global DEBUG, LOG_LEVEL, API_BASE, API_CACHE_TIME, CACHE_ADAPTER
    rz = dict(
        DEBUG='true' if settings.DEBUG else 'false',
        LOG_LEVEL=str(settings.LOG_LEVEL),
        API_BASE=settings.API_BASE,
        API_CACHE_TIME=str(settings.API_CACHE_TIME),
        CACHE_ADAPTER=settings.CACHE_ADAPTER,
        DEFAULT_DP=str(settings.CONF.dp),
        BASE_COIN=settings.CONF.base_coin,
        CURRENCY_LIST=','.join(settings.CONF.cur_list),
        CUR_SYMBOLS=to_keyval(settings.CUR_SYMBOLS),
    )
    return {**rz, **dict(kwargs)}


def gen_env(**kwargs) -> str:
    dz = _gen_env(**kwargs)
    res = [f"{k}={v}" for k, v in dz.items()]
    return "\n".join(res) + "\n"


@dataclass
class Pair(DictDataClass):
    from_sym: str
    to_sym: str
    rate: Decimal = field(default_factory=lambda: Decimal('0'))
    
    @property
    def pair(self) -> str:
        return f"{self.from_sym.upper()}_{self.to_sym.upper()}"
    
    def __post_init__(self):
        if not isinstance(self.rate, Decimal):
            self.rate = Decimal(self.rate)
        self.from_sym = self.from_sym.upper()
        self.to_sym = self.to_sym.upper()
    
    def is_pair(self, pair: str) -> bool:
        from_sym = self.from_sym.upper()
        to_sym = self.to_sym.upper()
        if pair.upper() == f"{from_sym}{to_sym}": return True
        if pair.upper() == f"{from_sym}_{to_sym}": return True
        if pair.upper() == f"{from_sym}/{to_sym}": return True
        if pair.upper() == f"{from_sym}-{to_sym}": return True
        return False
    
    def is_pair_sym(self, from_sym: str, to_sym: str) -> bool:
        return self.is_pair(f"{from_sym.upper()}_{to_sym.upper()}")
    
    def invert(self) -> "Pair":
        return Pair(from_sym=self.to_sym, to_sym=self.from_sym, rate=Decimal('1') / self.rate)
    
    def __str__(self):
        return str(self.rate)
    
    def __int__(self):
        return int(str(self))
    
    def __float__(self):
        return float(str(self))
    
    def __lt__(self, other):
        if isinstance(other, (str, float, int, Decimal)):
            return self.rate < Decimal(other)
        if isinstance(other, Pair):
            return self.rate < other.rate
        return False

    def __gt__(self, other):
        if isinstance(other, (str, float, int, Decimal)):
            return self.rate > Decimal(other)
        if isinstance(other, Pair):
            return self.rate > other.rate
        return False

    def __eq__(self, other):
        if isinstance(other, str):
            return self.is_pair(other) or other == str(self)
        if isinstance(other, (float, int, Decimal)):
            return Decimal(other) == self.rate
        if isinstance(other, Pair):
            if self.from_sym.upper() != other.from_sym.upper(): return False
            if self.to_sym.upper() != other.to_sym.upper(): return False
            if self.rate != other.rate: return False
            return True
        return False
    
    def __contains__(self, item):
        item = str(item)
        if item == str(self): return True
        if item.upper() == self.from_sym.upper(): return True
        if item.upper() == self.to_sym.upper(): return True
        if self.is_pair(item): return True
        return False


AnyPair = Union[Pair, "PairList"]
IterAnyPair = Union[List[AnyPair], Tuple[AnyPair, ...]]


@dataclass
class PairList(DictDataClass):
    """
    Represents a list of :class:`.Pair` objects, and provides magic methods to make it easy to query the list
    of pairs painlessly, as if the pair list were a dictionary.
    
    Examples::
    
        >>> pl = PairList([Pair('EUR', 'USD', Decimal('1.1986'))])
        >>> pl += Pair('EUR', 'GBP', Decimal('0.86793'))
        >>> 'eurgbp' in pl
        True
        >>> 'EUR/USD' in pl
        True
        >>> pl['eurgbp']
        Pair(from_sym='EUR', to_sym='GBP', rate=Decimal('0.86793'))
        >>> pl['EUR/GBP']
        Pair(from_sym='EUR', to_sym='GBP', rate=Decimal('0.86793'))
        >>> pl.find('eur', 'usd')
        Pair(from_sym='EUR', to_sym='USD', rate=Decimal('1.1986'))
        >>> pl['eur_gbp'].invert()
        Pair(from_sym='GBP', to_sym='EUR', rate=Decimal('1.152166649384166925904162778'))
        >>> pl.filter('eur')
        [Pair(from_sym='EUR', to_sym='USD', rate=Decimal('1.1986')),
         Pair(from_sym='EUR', to_sym='GBP', rate=Decimal('0.86793'))]
    
    """
    pairs: List[Pair] = field(default_factory=list)
    
    def find(self, from_sym: str, to_sym: str) -> Optional[Pair]:
        for p in self.pairs:
            if p.is_pair_sym(from_sym, to_sym): return p
        return None

    def find_pair(self, pair: str) -> Optional[Pair]:
        for p in self.pairs:
            if p.is_pair(pair): return p
        return None
    
    def filter(self, from_sym: str = None, to_sym: str = None) -> List[Pair]:
        if not empty(from_sym) and not empty(to_sym): return [p for p in self.pairs if p.is_pair_sym(from_sym, to_sym)]
        if not empty(from_sym): return [p for p in self.pairs if p.from_sym == from_sym.upper()]
        if not empty(to_sym): return [p for p in self.pairs if p.to_sym == to_sym.upper()]
        return []
    
    def add(self, other: Union[AnyPair, IterAnyPair]) -> "PairList":
        if isinstance(other, (PairList, Pair)): return self._add(other)
        if isinstance(other, (list, tuple)): return self._merge(other)
        raise ValueError(f"Cannot add object of type {type(other)} (repr: {other!r}) to PairList")
    
    append = add

    def _add(self, other: AnyPair) -> "PairList":
        if isinstance(other, PairList): self.pairs += other.pairs
        elif isinstance(other, Pair): self.pairs.append(other)
        else: raise ValueError(f"Cannot add object of type {type(other)} (repr: {other!r}) to PairList")
        return self
    
    def _merge(self, other: IterAnyPair) -> "PairList":
        if len(other) > 0:
            if isinstance(other[0], Pair):
                self.pairs += list(other)
            elif isinstance(other[0], PairList):
                for pl in other:
                    self.pairs += pl.pairs
            else:
                raise ValueError(f"Cannot add object of type {type(other)} (repr: {other!r}) to PairList")
        return self
    
    def __getitem__(self, item) -> Union[Pair, List[Pair]]:
        if item == 'pairs':
            return self.pairs
        
        xpair = self.find_pair(item)
        if not empty(xpair): return xpair
        
        filtered = self.filter(item)
        if len(filtered) > 0: return filtered
        raise KeyError(f"Key/pair '{item}' not found.")
    
    def __add__(self, other: Union[AnyPair, IterAnyPair]) -> "PairList":
        return self.add(other)
    
    def __contains__(self, item):
        if not empty(self.find_pair(item), itr=True): return True
        if not empty(self.filter(item), itr=True): return True
        if not empty(self.filter(to_sym=item), itr=True): return True
        for p in self.pairs:
            if p == item: return True
        return False
    

class CurrencyAdapter(abc.ABC):
    """
    This is the base class for all privex-curconv currency adapters, which includes various functions pre-scaffolded,
    allowing a child class to simply implement :meth:`._get_rates` and ideally change :attr:`.DEFAULT_CACHE_PREFIX`
    to be fully functional.
    
    It handles higher level functionality such as caching and some common configuration options, so
    that child classes only need to worry about obtaining the exchange rates, and converting them
    into a :class:`.PairList` / :class:`.Pair`
    
    An adapter implementing this class can be used like so::
        
        >>> rp = RatesAPIAdapter()
        >>> res = await rp.get_rates()
        >>> 'eurusd' in res
        True
        >>> res['eurusd']
        Pair(from_sym='EUR', to_sym='USD', rate=Decimal('1.1986'))
    
    
    """
    DEFAULT_CACHE_PREFIX: str = "conv:get_rates"
    CTX_SLEEP_PRE: float = 0.1
    CTX_SLEEP_POST: float = 0.1
    _rates: Optional[PairList]
    _rates_timeout: Optional[int]
    rates_refresh: int
    api_base: str
    api_uri: str
    httpx_conf: dict
    base_coin: str
    cache: Union[AsyncCacheAdapter, AsyncCacheWrapper]
    cache_timeout: Union[float, int, Decimal]
    cache_prefix: str
    _http: Optional[httpx.AsyncClient]
    
    def __init__(self, base_coin: Optional[str] = None, **kwargs):
        self.api_base = kwargs.get('api_base', '')
        self.api_uri = kwargs.get('api_uri', '')
        self.httpx_conf = kwargs.get('httpx_conf', {})
        self.base_coin = base_coin
        self._rates = None
        self._rates_timeout = None
        self._http = None
        
        self.cache = kwargs.get('cache', async_cached)
        self.cache_timeout = kwargs.get('cache_timeout', settings.API_CACHE_TIME)
        self.cache_prefix = kwargs.get('cache_prefix', self.DEFAULT_CACHE_PREFIX)
        
        self.rates_refresh = int(kwargs.get('rates_refresh', self.cache_timeout))
    
    def _make_httpx(self, **kwargs) -> httpx.AsyncClient:
        hconf = {k: v for k, v in self.httpx_conf.items()}
        hconf = {**hconf, **kwargs}
        return httpx.AsyncClient(**hconf)
    
    @property
    def httpx(self) -> Union[httpx.AsyncClient, AsyncContextManager]:
        if self._http:
            log.debug(f"Detected {self.__class__.__name__!s}._http - returning http_wrapper for re-using client")
    
            @contextlib.asynccontextmanager
            async def http_wrapper() -> AsyncContextManager:
                log.debug(f"Yielding {self.__class__.__name__!s}._http from http_wrapper")
                yield self._http
                log.debug(f"Returned from yielding {self.__class__.__name__!s}._http in http_wrapper")

            return http_wrapper()
        log.debug(f"No {self.__class__.__name__!s}._http - returning throwaway AsyncClient")
        return self._make_httpx()
    
    @abc.abstractmethod
    async def _get_rates(self, base: str = None, *symbols, **kwargs) -> PairList:
        raise NotImplementedError(f"{self.__class__.__name__} must implement _get_rates!")
    
    async def _get_rate(self, from_sym: str = None, to_sym: str = None, **kwargs) -> Pair:
        r: PairList = await self.rates
        return r.find(from_sym, to_sym)
    
    def _get_prefix(self, *args, _extra_prefix="", **kwargs) -> str:
        return f"{self.cache_prefix}:{_extra_prefix}{dump_args(*args, **kwargs)}"

    async def get_rate(self, from_sym: str = None, to_sym: str = None, **kwargs) -> Pair:
        cpfx = self._get_prefix(from_sym=str(from_sym).upper(), to_sym=str(to_sym).upper(), _extra_prefix='rate:', **kwargs)
        res = await await_if_needed(self.cache.get, cpfx)
        if empty(res, itr=True, zero=True):
            res = await await_if_needed(self._get_rate, from_sym, to_sym, **kwargs)
            await await_if_needed(self.cache.set, cpfx, res, self.cache_timeout)
        return res

    async def get_rates(self, base: str = None, *symbols, reset=False, **kwargs) -> PairList:
        # Unless ._rates is valid and not expired, then we need to try the external cache, or query directly.
        if reset or self.rates_expired:
            log.debug(f" [{self.__class__.__name__}.get_rates] Rates not loaded / expired. Reading cache.")
            cpfx = self._get_prefix(base=str(base).upper(), symbols=sorted(list(symbols)), **kwargs)
            # Try loading rates from the cache. If there's no valid rates in the cache, then call .do_get_rates
            # to get fresh rates + set both the local _rates cache and external cache.
            res = await await_if_needed(self.cache.get, cpfx)
            if reset or empty(res, itr=True, zero=True):
                log.debug(f" [{self.__class__.__name__}.get_rates] Rates not in cache - calling HTTP(S) API")
                res = await self.do_get_rates(base, *symbols, **kwargs)
            return res
        log.debug(f" [{self.__class__.__name__}.get_rates] Rates already in memory. Using in-memory rates: {self._rates}")
        return self._rates

    async def do_get_rates(self, base: str = None, *symbols, set_cache=True, **kwargs):
        cpfx = self._get_prefix(base=str(base).upper(), symbols=sorted(list(symbols)), **kwargs)
        log.debug(f" [{self.__class__.__name__}.do_get_rates] Calling API via ._get_rates")

        res = await await_if_needed(self._get_rates, base, *symbols, **kwargs)
        if set_cache and not empty(res, itr=True):
            log.debug(f" [{self.__class__.__name__}.do_get_rates] Storing results via privex.helpers async cache")
            await await_if_needed(self.cache.set, cpfx, res, timeout=self.cache_timeout)
            self._set_rates(res)
        return res

    @property
    def rates_expired(self) -> bool:
        if empty(self._rates, itr=True, zero=True): return True
        if empty(self._rates_timeout, zero=True): return True
        return not (int(time.time()) < int(self._rates_timeout))
    
    @async_property
    async def rates(self) -> PairList:
        if self.rates_expired:
            return await self.get_rates()
        return self._rates
    
    def _set_rates(self, rates: PairList) -> PairList:
        self._rates = rates
        self._rates_timeout = int(time.time()) + int(self.rates_refresh)
        return self._rates
    
    async def trans_rate(self, from_coin: str, to_coin: str) -> Decimal:
        pass
    
    async def __aenter__(self):
        log.debug(f" [{self.__class__.__name__}.__aenter__] Creating and storing httpx client")
        self._http = self._make_httpx()
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._http is not None:
            log.debug(f" [{self.__class__.__name__}.__aexit__] Closing httpx client")
            try:
                if self.CTX_SLEEP_PRE > 0: await asyncio.sleep(self.CTX_SLEEP_PRE)
                await self._http.aclose()
            except Exception as e:
                log.warning(f"Failed to close httpx.AsyncClient. Reason: {type(e)} - {e!s}")
            finally:
                if self.CTX_SLEEP_POST > 0: await asyncio.sleep(self.CTX_SLEEP_POST)
                self._http = None


def trans_rate(from_coin: str, to_coin: str, rates: PairList, base_coin: str = settings.CONF.base_coin) -> Decimal:
    """
    Get a translated rate using ``rates``

    Example::

        >>> rates = await get_rates("EUR")
        >>> trans_rate("USD", "SEK", rates)
        Decimal('8.556303086367841224455470524')
        >>> trans_rate("SEK", "USD", rates)
        Decimal('0.1168729052612956173889112764')

    """
    if from_coin.upper() == to_coin.upper(): return Decimal('1.000000')
    if 'rates' in rates and isinstance(rates.get('rates', None), dict):
        rates = rates['rates']
    # from_rate = rates.get(from_coin, rates.get(from_coin.upper()))
    # to_rate = rates.get(to_coin, rates.get(to_coin.upper()))
    from_rate = rates.find(base_coin, from_coin)
    to_rate = rates.find(base_coin, to_coin)
    if from_rate is None: raise AttributeError(f"from_coin {from_coin!r} was not found in 'rates'!")
    if to_rate is None: raise AttributeError(f"to_coin {to_coin!r} was not found in 'rates'!")
    return (Decimal('1') / from_rate.rate) * to_rate.rate


def trans_dict(from_coin: str, rates: PairList) -> Dict[str, Decimal]:
    # if 'rates' in rates and isinstance(rates.get('rates', None), dict):
    #     rates = rates['rates']
    # return {k: trans_rate(from_coin, k, rates) for k in list(rates.keys())}
    return {p.to_sym: trans_rate(from_coin, p.to_sym, rates) for p in rates.pairs}


def is_numstr(d: str):
    if len(d) == 0: return False
    if not d[0].isdigit() and not d[0] in ['.', '-', '+']:
        return False
    hasdot, has_e = False, False
    for c in d[1:]:
        if c == '.':
            if hasdot: return False
            hasdot = True
            continue
        if c == 'e':
            if has_e: return False
            has_e = True
            continue
        if c.isdigit() or c in [',', '-', '+']:
            continue
        return False
    return True

